//
//  GLTFOpenGLView.m
//
//  gl_triangle_fan
//  Illusioner
//
//  Created by Tom Bernard on 10/1/06.
//  Copyright (c) 2006, Bersearch Information Services. All rights reserved.
//

#import "GLTFOpenGLView.h"
#import "Benchmarker.h"

extern void printfNSRect(char *label, NSRect rect);

void hsbaToRgba(float _hue, float _saturation, float _brightness, float _hsbAlpha,
				float *_red, float *_green, float *_blue, float *_rgbAlpha)
{
	NSColor * color = [NSColor colorWithCalibratedHue:_hue saturation:_saturation brightness:_brightness alpha:_hsbAlpha];
	
	*_red = [color redComponent];
	*_green = [color greenComponent];
	*_blue = [color blueComponent];
	*_rgbAlpha = [color alphaComponent];
}

/*
float randomf()
 
 produces a random number between 0.0 and 1.0
 */

float randomf()
{
	return (float)random() / (float)(pow(2, 31) - 1);
}

/*
 void renderCircleAsPolygon(float originx, float originy, float r, float n)
 
 renderCircleAsPolygon renders a circle at originx, originy with a radius of r, 
 approximating the circle with n sides
 */

void renderCircleAsPolygon(float originx, float originy, float r, float n)
{
	float x, y, i, pi, pi2;
	
	pi = acos(-1.0);
	pi2 = 2 * pi;
	
	glBegin(GL_POLYGON);
	for (i = 0.0; i <= n; i++)
	{
		float theta = i * pi2 / n;
		x = r * cos(theta) + originx;
		y = r * sin(theta) + originy;
		glVertex3f(x, y, 0.0);
	}
	glEnd();
	
}


void reshapeGL(int x, int y, int width, int height)
{
	glViewport(x, y, width, height);
}


@implementation GLTFOpenGLView

- (id)initWithFrame:(NSRect)frameRect
		pixelFormat:(NSOpenGLPixelFormat*)format
{
// 	const double millisecondsMultiplier = 0.001;
	const double microsecondsMultiplier = 1.0;
// 	const double nanosecondsMultiplier = 1000.0;
// 	const double noResultsRequested = 0.0;
	
	
    self = [super initWithFrame:frameRect pixelFormat:format];
	if (self)
	{
		// inited = 0;
		
		drawRectBenchmark = [[Benchmarker alloc] initBenchmarkerForSamples:100
														  outputData:NO
													   outputResults:YES
												  withUnitMultiplier:microsecondsMultiplier
												 disposeDataWhenDone:YES
												  nsLogStampStartEnd:YES
														   userLabel:@"-drawRect:"];
		
		
		pi = acos(-1.0);
		pi2 = 2 * pi;
		
		hue = 0.5;
		saturation = 1.0;
		brightness = 1.0;
		hsbAlpha = 1.0;
		
		baseHue = hue;
		
		hsbaToRgba(hue, saturation, brightness, hsbAlpha, &red, &green, &blue, &rgbAlpha);
		
		upperD_addedHueFactor = 20.0;
		lowerD_addedHueFactor = 1.0;
		// reality check; not needed yet, but prepare for the future...
		if (upperD_addedHueFactor < lowerD_addedHueFactor)
		{
			float swap = upperD_addedHueFactor;
			upperD_addedHueFactor = lowerD_addedHueFactor;
			lowerD_addedHueFactor = swap;
		}
		dd_addedHueFactor = -0.05;
		d_addedHueFactor = upperD_addedHueFactor;
		
		sSides = 3.0;
		eSides = 32.0;
		dSides = 0.1;
		sides = sSides;
		
		lowerOrthoXdelta = 1.0;
		upperOrthoXdelta = 10.0;
		// reality check
		if (lowerOrthoXdelta > upperOrthoXdelta)
		{
			GLfloat swap = lowerOrthoXdelta;
			lowerOrthoXdelta = upperOrthoXdelta;
			upperOrthoXdelta = swap;
		}
		dOrthoXdelta = 0.1;
		orthoXdelta = lowerOrthoXdelta;
		
		lowerOrthoYdelta = 1.0;
		upperOrthoYdelta = 10.0;
		// reality check
		if (lowerOrthoYdelta > upperOrthoYdelta)
		{
			GLfloat swap = lowerOrthoYdelta;
			lowerOrthoYdelta = upperOrthoYdelta;
			upperOrthoYdelta = swap;
		}
		dOrthoYdelta = 0.1;
		orthoYdelta = lowerOrthoYdelta;
		
		sOrthoZdelta = 1.0;
		eOrthoZdelta = 10.0;
		dOrthoZdelta = 0.1;
		orthoZdelta = sOrthoZdelta;
		
		[self calculateOrthoXYZStartEnd];
		
		firstPointMode = firstPointOnCircle;
	}
	return self;
	
}

- (id)initWithFrame:(NSRect)frameRect
{
	// printf("-[%s initWithFrame:] ", [[self description] UTF8String]);
	// printfNSRect("frameRect", frameRect);
	
	
	NSOpenGLPixelFormatAttribute attrs[] =
	{
		NSOpenGLPFADoubleBuffer,
		NSOpenGLPFADepthSize, (NSOpenGLPixelFormatAttribute)32,
		(NSOpenGLPixelFormatAttribute)nil
	};
	
	NSOpenGLPixelFormat* pixFmt = [[[NSOpenGLPixelFormat alloc] initWithAttributes:attrs] autorelease];
	
	self = [self initWithFrame:frameRect pixelFormat:pixFmt];
	return self;
}

- (void)dealloc
{
	[drawRectBenchmark release];
	[super dealloc];
}


- (void)drawRect:(NSRect)rect
{
	/*
	printf("-[%s drawRect:] ", [[self description] UTF8String]);
	printfNSRect("rect", rect);
	NSWindow *myWindow = [[self superview] window];
	printf("%s ", [[myWindow description] UTF8String]);
	printfNSRect("frame ", [myWindow frame]);
	 */
	
	[drawRectBenchmark startBenchmark];
	
	if (!inited) [self delayedInit];
	
	glClear(GL_COLOR_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// printf("ortho = %g\n", ortho);
	// glOrtho(0.0, ortho, 0.0, ortho, 0.0, ortho);
	glOrtho(orthoXstart, orthoXend, orthoYstart, orthoYend, orthoZstart, orthoZend);
	glMatrixMode(GL_MODELVIEW);
	
	float x, y;
	float addedHue = 0.0;
	 float workingSides, sidesShift = 0.0;

	 // more complex color animation using GL_TRIANGLE_FAN
	 for (x = 0.05; x < 1.0; x += 0.1)	
	 {
		 for (y = 0.05; y < 1.0; y += 0.1)
		 {
			 workingSides = sides + sidesShift;
			 if (workingSides > eSides) workingSides -= eSides;
			 
			 sidesShift += 2.0 * dSides;
			 if (sidesShift > (eSides - sSides)) sidesShift = 0.0;
			 
			 hue = baseHue + addedHue;
			 addedHue += 0.01;
			 if (hue > 1.0) hue -= 1.0;
			 hsbaToRgba(hue, saturation, brightness, hsbAlpha, &red, &green, &blue, &rgbAlpha);
			 
			 glColor4f(red, green, blue, rgbAlpha);
			 			 
			 [self renderCircleAsTriangleFanAtX:x
											  y:y
									 withRadius:0.05
									   andSides:workingSides];
			 			 
			 [self renderColorfulShapeAtX:x 
										y:y 
							   withRadius:0.045
									sides:workingSides
									  hue:hue
								  andMode:GL_TRIANGLE_FAN];
		 }
	 }
	
    [[self openGLContext] flushBuffer];
	//  glutSwapBuffers();
	
	[drawRectBenchmark endBenchmark];
}


- (void)animate:(NSTimer *)timer
{
	// printf("-[%s animate:]\n", [[self description] UTF8String]);
	baseHue += 0.01;
	if (baseHue > 1.0) baseHue = 0.0;
	
	if (addedHueFactorMode)
	{
		d_addedHueFactor += dd_addedHueFactor;
		if (d_addedHueFactor < lowerD_addedHueFactor || d_addedHueFactor > upperD_addedHueFactor)
		{
			if (dd_addedHueFactor < 0.0)
			{
				d_addedHueFactor = upperD_addedHueFactor;
			}
			else
			{
				d_addedHueFactor = lowerD_addedHueFactor;
			}
		}
	}
	
	sides += dSides;
	if (sides > eSides) sides = sSides;
	
	orthoXdelta += dOrthoXdelta;
	if (orthoXdelta < lowerOrthoXdelta || orthoXdelta > upperOrthoXdelta)
	{
		if (dOrthoXdelta < 0.0)
		{
			orthoXdelta = upperOrthoXdelta;
		}
		else
		{
			orthoXdelta = lowerOrthoXdelta;
		}
	}
	
	orthoYdelta += dOrthoYdelta;
	if (orthoYdelta < lowerOrthoYdelta || orthoYdelta > upperOrthoYdelta)
	{
		if (dOrthoYdelta < 0.0)
		{
			orthoYdelta = upperOrthoYdelta;
		}
		else
		{
			orthoYdelta = lowerOrthoYdelta;
		}
	}
	
	
	
	/*
	 orthoZdelta += dOrthoZdelta;
	 if (orthoZdelta > eOrthoZdelta) orthoZdelta = sOrthoZdelta;
	 */
	
	[self calculateOrthoXYZStartEnd];
	[self setNeedsDisplay:YES];
}

- (void)delayedInit
{
	if(!inited)
	{
        [[self openGLContext] makeCurrentContext];
        
		swap_enabled = 1;
		[[self openGLContext] setValues:&swap_enabled forParameter:NSOpenGLCPSwapInterval];
		
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
        
        inited = YES;
		
		glMatrixMode(GL_MODELVIEW);
		
		[self reshape];
    }
	
}

- (void)reshape
{
    if(inited)
    {
		[[self openGLContext] makeCurrentContext];
		
		NSRect frame = [self frame];
		reshapeGL((int)frame.origin.x, (int)frame.origin.y, (int)frame.size.width, (int)frame.size.height);
		[self setNeedsDisplay:YES];
    }
}

- (void)renderCircleAsTriangleFanAtX:(float)originx
								   y:(float)originy
						  withRadius:(float)r
							andSides:(float)n
{
	float x, y, i;
	
	glBegin(GL_TRIANGLE_FAN);
	for (i = 0.0; i <= n; i++)
	{
		
		float theta = i * pi2 / n;
		x = r * cos(theta) + originx;
		y = r * sin(theta) + originy;
		glVertex3f(x, y, 0.0);
	}
	glEnd();
	
}

- (void)renderColorfulShapeAtX:(float)originx 
							 y:(float)originy 
					withRadius:(float)r
						 sides:(float)n
						   hue:(float)_baseHue
					   andMode:(GLenum)mode
{
	float x, y, i, _hue, _addedHue = 0, d_addedHue, _red, _green, _blue, _rgbalpha;
	
	if (addedHueFactorMode)
	{
		d_addedHue = d_addedHueFactor / n;
	}
	else
	{
		d_addedHue = 1.0 / n;
	}
	
	glBegin(mode);
	
	/*
	 switch(firstPointMode)
	 {
		 case zeroZero:
			 glVertex3f(0.0, 0.0, 0.0);
			 break;
			 
		 case originOrigin:
			 glVertex3f(originx, originy, 0.0);
			 break;
	 }
	 */
	
	if (firstPointMode == zeroZero) glVertex3f(0.0, 0.0, 0.0);
	if (firstPointMode == originOrigin) glVertex3f(originx, originy, 0.0);
	if (firstPointMode == randomFirstPoint) [self doRandomFirstPointWithRadius:r
																		   atX:originx
																			 y:originy];
	
	
	for (i = 0.0; i <= n; i++)
	{
		float theta = i * pi2 / n;
		x = r * cos(theta) + originx;
		y = r * sin(theta) + originy;
		
		_hue = _baseHue + _addedHue;
		_addedHue += d_addedHue;
		if (_hue > 1.0) _hue -= 1.0;
		hsbaToRgba(_hue, 1.0, 1.0, 1.0, &_red, &_green, &_blue, &_rgbalpha);
		
		glColor4f(_red, _green, _blue, _rgbalpha);
		glVertex3f(x, y, 0.0);
	}
	
	glEnd();
	
}

- (void)doRandomFirstPointWithRadius:(GLfloat)r 
								 atX:(GLfloat)originx
								   y:(GLfloat)originy
{
	GLfloat x, y;
	
	x = r * (2.0 * randomf() - 1.0) + originx;
	y = r * (2.0 * randomf() - 1.0) + originy;
	glVertex3f(x, y, 0.0);
}

- (void)negateDOrthoXdelta
{
	dOrthoXdelta *= -1.0;
}

- (void)negateDOrthoYdelta
{
	dOrthoYdelta *= -1.0;
}

- (void)negateDOrthoZdelta
{
	dOrthoZdelta *= -1.0;
}


- (void)calculateOrthoXYZStartEnd
{
	orthoXstart = 0.5 - 0.5 * orthoXdelta;
	orthoXend = orthoXstart + orthoXdelta;
	
	orthoYstart = 0.5 - 0.5 * orthoYdelta;
	orthoYend = orthoYstart + orthoYdelta;
	
	orthoZstart = 0.5 - 0.5 * orthoZdelta;
	orthoZend = orthoZstart + orthoZdelta;
}

- (void)toggleAddedHueFactorMode
{
	addedHueFactorMode = !addedHueFactorMode;
}

- (void)incrementFirstPointMode
{
	firstPointMode += 1;
	if (firstPointMode > firstPointOnCircle) firstPointMode = zeroZero;
}


@end

