//
//  gl_triangle_fanView.m
//
//  gl_triangle_fan
//  Illusioner
//
//  Created by Tom Bernard on 10/1/06.
//  Copyright (c) 2006, Bersearch Information Services. All rights reserved.
//

#import "gl_triangle_fanView.h"
#import "GLTFOpenGLView.h"


void printfNSRect(char *label, NSRect rect)
{
	printf("%s: x = %g, y = %g, w = %g, h = %g\n", label, rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
}

@implementation gl_triangle_fanView

- (id)initWithFrame:(NSRect)frame isPreview:(BOOL)isPreview
{
	self = [super initWithFrame:frame isPreview:isPreview];
	if (self)
	{
		animationTimeInterval = 1.0 / 60.0;
		[self setAnimationTimeInterval:animationTimeInterval];
		
		GLTFOpenGLView *gltfOpenGLView = [[[GLTFOpenGLView alloc] initWithFrame:NSMakeRect(0.0, 0.0, frame.size.width, frame.size.height)] autorelease];
		if (gltfOpenGLView) [self addSubview:gltfOpenGLView];
	}
    return self;
}


- (void)dealloc
{
	// [gltfOpenGLView release];
	[super dealloc];
}

- (void)startAnimation
{
    [super startAnimation];
}

- (void)stopAnimation
{
    [super stopAnimation];
}

- (void)drawRect:(NSRect)rect
{
   // [super drawRect:rect];
	
	printf("- [%s drawRect:] ", [[self description] UTF8String]);
	printfNSRect("rect", rect);
	printf("%s\n", [[[self window] description] UTF8String]);
}

- (void)animateOneFrame
{
	[self animateOpenGLSubviews];
	// [self setAnimationTimeInterval:[self animationTimeInterval]];
}

- (BOOL)hasConfigureSheet
{
    return NO;
}

- (NSWindow*)configureSheet
{
    return nil;
}

- (void)animateOpenGLSubviews
{
	NSArray *subviews = [self subviews];
	id aView;
	NSEnumerator *en = [subviews objectEnumerator];
	while (aView = [en nextObject])
	{
		// printf("%s\n", [[aView description] UTF8String]);
		if ([aView isKindOfClass:[GLTFOpenGLView class]]) [((GLTFOpenGLView *)aView) animate:nil];
	}
}

/*
- (NSWindow *)window
{
	NSWindow *window = [super window];
	if (!window)
	{
		NSView *currentView = self;
		currentView = [currentView superview];
		// printf(" \n");
		while (currentView)
		{
			window = [currentView window];
			if (window) return window;
			currentView = [currentView superview];
		}
	}
	return window;
}
 */


- (NSTimeInterval)animationTimeInterval
{
	animationTimeInterval += dAnimationRate;
	if (animationTimeInterval > eAnimationRate) animationTimeInterval = sAnimationRate;
	if (animationTimeInterval < sAnimationRate) animationTimeInterval = eAnimationRate;
	// printf("animationTimeInterval = %g\n", animationTimeInterval);
	return animationTimeInterval;
}

// pressing a modifier key kills ScreenSaverEngine before our -keyDown: is called. No point in testing for modifiers.

/*
- (void)printfModifierFlags
{
	if (optionKeyDown)		printf("optionKeyDown ");
	if (controlKeyDown)		printf("controlKeyDown ");
	if (commandKeyDown)		printf("commandKeyDown ");
	if (shiftDown)			printf("shiftDown ");
	if (capsLockDown)		printf("capsLockDown ");
	printf("pressed\n");
}

- (void)setFlagsfromEvent:(NSEvent *)theEvent
{
	unsigned int modifiers = [theEvent modifierFlags];
		
	optionKeyDown =			NSAlternateKeyMask == (modifiers & NSAlternateKeyMask);
	controlKeyDown =		NSControlKeyMask == (modifiers & NSControlKeyMask);
	commandKeyDown =		NSCommandKeyMask == (modifiers & NSCommandKeyMask);
	shiftDown =				NSShiftKeyMask == (modifiers & NSShiftKeyMask);
	capsLockDown =			NSAlphaShiftKeyMask == (modifiers & NSAlphaShiftKeyMask);
	
	[self printfModifierFlags];
}
 */


- (void)keyDown:(NSEvent *)theEvent
{
	// printf("-keyDown:\n");
	NSString *characters = [theEvent charactersIgnoringModifiers];
	unichar firstCharacter = [characters characterAtIndex:0];
	
	// [self setFlagsfromEvent:theEvent];
	
	switch (firstCharacter)
	{
		case ' ':
			[self doCaseSpace];				// toggle animation
			break;
			
		case 'm':							// change hue animation
			[self doCasem];
			break;
			
		case 'e':
			[self doCasee];
			break;
			
		case 'p':							// change first point mode
			[self doCasep];
			break;
			
		case 'r':							// renew resumeAnimationTimer
			[self doCaser];
			break;
			
		case 's':							// single step one frame
			[self doCases];
			break;
			
		case 'x':							// negate delta ortho x
			[self doCasex];
			break;
			
		case 'y':							// negate delta ortho y
			[self doCasey];
			break;
			
		case 'z':							// negate delta ortho z
			[self doCasez];
			break;
			
		default:
			[super keyDown:theEvent];		// quit ScreenSaverEngine
	}
}

- (void)doCasem
{
	[self toggleAddedHueFactorModeOfGLSubviews];
}

- (void)doCasee
{
	
}

- (void)doCasep
{
	[self incrementFirstPointModeOfGLSubviews];
}

- (void)doCaser
{
	[self renewResumeAnimationTimer];
}

- (void)doCases
{
	[self doSingleStep];
}

- (void)doCasex
{
	[self negateDOrthoXdeltaOfSubviews];
}

- (void)doCasey
{
	[self negateDOrthoYdeltaOfSubviews];
}

- (void)doCasez
{
	[self negateDOrthoZdeltaOfSubviews];
}

- (void)doCaseSpace
{
	[self toggleAnimationState];
}

- (void)incrementFirstPointModeOfGLSubviews
{
	NSArray *subviews = [self subviews];
	id aView;
	NSEnumerator *en = [subviews objectEnumerator];
	while (aView = [en nextObject])
	{
		if ([aView isKindOfClass:[GLTFOpenGLView class]]) [((GLTFOpenGLView *)aView) incrementFirstPointMode];
	}
	[self setNeedsDisplay:YES];
}

- (void)negateDOrthoXdeltaOfSubviews
{
	NSArray *subviews = [self subviews];
	id aView;
	NSEnumerator *en = [subviews objectEnumerator];
	while (aView = [en nextObject])
	{
		if ([aView isKindOfClass:[GLTFOpenGLView class]]) [((GLTFOpenGLView *)aView) negateDOrthoXdelta];
	}
}


- (void)negateDOrthoYdeltaOfSubviews
{
	NSArray *subviews = [self subviews];
	id aView;
	NSEnumerator *en = [subviews objectEnumerator];
	while (aView = [en nextObject])
	{
		if ([aView isKindOfClass:[GLTFOpenGLView class]]) [((GLTFOpenGLView *)aView) negateDOrthoYdelta];
	}
}

- (void)negateDOrthoZdeltaOfSubviews
{
	
}

- (void)toggleAddedHueFactorModeOfGLSubviews
{
	NSArray *subviews = [self subviews];
	id aView;
	NSEnumerator *en = [subviews objectEnumerator];
	while (aView = [en nextObject])
	{
		// printf("%s\n", [[aView description] UTF8String]);
		if ([aView isKindOfClass:[GLTFOpenGLView class]]) [((GLTFOpenGLView *)aView) toggleAddedHueFactorMode];
	}
	[self setNeedsDisplay:YES];
}

- (void)toggleAnimationState
{
	if ([self isAnimating])
	{
		[self stopAnimation];
		[self newResumeAnimationTimer];
	}
	else
	{
		[self startAnimation];
	}
}

- (void)doSingleStep
{
	if ([self isAnimating]) [self stopAnimation];
	[self newResumeAnimationTimer];
	[self animateOneFrame];
}

- (void)renewResumeAnimationTimer
{
	[self newResumeAnimationTimer];
}

- (void)newResumeAnimationTimer
{
	if (resumeAnimationTimer) [resumeAnimationTimer invalidate];
	resumeAnimationTimer = [NSTimer scheduledTimerWithTimeInterval:kResumeAnimationTime
															 target:self
														   selector:@selector(resumeAnimationForTimer:)
														   userInfo:nil
															repeats:NO];
}

- (void)resumeAnimationForTimer:(NSTimer *)timer
{
	resumeAnimationTimer = nil;
	if (![self isAnimating]) [self startAnimation];
}

@end
